package org.qcg.broker.schemas.hostparams.wrapper;

import org.qcg.broker.schemas.exception.NoSuchParamException;
import org.qcg.broker.schemas.exception.UnknownParameter;
import org.qcg.broker.schemas.hostparams.ComputingResource;
import org.qcg.broker.schemas.hostparams.HostParameters;
import org.qcg.broker.schemas.hostparams.QueueResource;
import org.qcg.broker.schemas.hostparams.wrapper.exception.NotHomogeneous;


/**
 * 
 * @author Marcin Krystek
 *
 */
public interface QueuingSystem {
	
	public void create(HostParameters hostParams, ComputingResource frontend);
	
	public ComputingResource getFrontend();
	
	/**
	 * 
	 * @return list of queues within queuing system.
	 * @throws UnknownParameter
	 */
	public QueueResource[] getQueues() throws UnknownParameter;
	
	/**
	 * 
	 * @return list of all computing nodes in queuing system. Computing node must have
	 * accurate reference to frontend and queue.
	 * @throws UnknownParameter
	 */
	public ComputingResource[] getComputingNodes() throws UnknownParameter;
	
	/**
	 * 
	 * @param queueResource
	 * @return return list of computing nodes which belongs to this queuing system and
	 * are available through specified queue.
	 * @throws UnknownParameter
	 */
	public ComputingResource[] getComputingNodes(QueueResource queueResource) throws UnknownParameter;

	/**
	 * 
	 * @return queuing system frontend id. This is the value of resourceId attribute
	 * in ComputingResource element.
	 */
	public String getFrontendId();

	/**
	 * 
	 * For all computing resources, which belongs to this queuing system, 
	 * calculates sum of {@link ComputingParameterName}.CPUCOUNT parameter values.
	 * @return total number of cpus available in this queuing system. 
	 * @throws NoSuchParamException if parameter is not defined.
	 * @throws UnknownParameter
	 */
	public double getCpuCnt() throws NoSuchParamException, UnknownParameter;

	/**
	 * For all computing resources, which belongs to this queuing system,
	 * calculates sum of {@link ComputingParameterName}.MEMORY parameter values.
	 * @return total memory available in this queuing system.
	 * @throws NoSuchParamException if parameter is not defined.
	 * @throws UnknownParameter
	 */
	public double getMemory() throws NoSuchParamException, UnknownParameter;
	
	/**
	 * For all computing resources, which belongs to this queuing system,
	 * calculates sum of {@link ComputingParameterName}.FREEMEMORY parameter values.
	 * @return total free memory available in this queuing system.
	 * @throws NoSuchParamException if parameter is not defined.
	 * @throws UnknownParameter
	 */
	public double getFreeMemory() throws NoSuchParamException, UnknownParameter;

	/**
	 * For all queues, which belongs to this queuing system, calculates sum of
	 * {@link QueueParameterName}.JOBSCOUNT parameter values.
	 * @return total number of jobs in queuing system.
	 * @throws NoSuchParamException if parameter is not defined.
	 * @throws UnknownParameter
	 */
	public double getJobsCount() throws NoSuchParamException, UnknownParameter;

	/**
	 * For all queues, which belongs to this queuing system, calculates sum of
	 * {@link QueueParameterName}.RUNNINGJOBSCOUNT parameter values.
	 * @return total number of jobs running in the queuing system.
	 * @throws NoSuchParamException if parameter is not defined.
	 * @throws UnknownParameter
	 */
	public double getRunningJobsCount() throws NoSuchParamException, UnknownParameter;
	
	public double getWaitingJobsCount() throws NoSuchParamException, UnknownParameter;

	/**
	 * For all computing resources, which belongs to this queuing system,
	 * calculates average value from stats values in {@link ComputingParameterName}.FREEMEMORY parameter.
	 * @return average free memory, available in this queuing system.
	 * @throws NoSuchParamException if parameter is not defined.
	 * @throws UnknownParameter
	 */
	public double getAvgFreeMemory() throws NoSuchParamException, UnknownParameter;
	
	/**
	 * For all computing resources, which belongs to this queuing system, calculates sum of
	 * {@link QueueParameterName}.FREECPUS parameter values.
	 * @return total count of free cpus available in queuing system.
	 * @throws NoSuchParamException if parameter is not defined.
	 * @throws UnknownParameter
	 */
	public double getFreeCpus() throws NoSuchParamException, UnknownParameter;

	/**
	 * For all computing resources, which belongs to this queuing system,
	 * calculates average value from stats values in {@link ComputingParameterName}.CPULOAD parameter.
	 * @return average load of cpus in queuing system.
	 * @throws NoSuchParamException if parameter is not defined.
	 * @throws UnknownParameter
	 */
	public double getAvgCpuLoad() throws NoSuchParamException, UnknownParameter;
	

	/**
	 * If queuing system is homogeneous, this method should return operating system name, common
	 * for all computing resources.
	 * @return operating system name.
	 * @throws NoSuchParamException if parameter is not defined.
	 * @throws UnknownParameter
	 * @throws NotHomogeneous if queuing system is not homogeneous. 
	 */
	public String getOsName() throws NoSuchParamException, UnknownParameter, NotHomogeneous;

	/**
	 * If queuing system is homogeneous, this method should return operating system release, common
	 * for all computing resources.
	 * @return operating system release.
	 * @throws NoSuchParamException if parameter is undefined.
	 * @throws UnknownParameter
	 * @throws NotHomogeneous if queuing system is not homogeneous.
	 */
	public String getOsRelease() throws NoSuchParamException, UnknownParameter, NotHomogeneous;

	/**
	 * If queuing system is homogeneous, this method should return operating system version, common
	 * for all computing resources.
	 * @return operating system version.
	 * @throws NoSuchParamException if parameter is undefined.
	 * @throws UnknownParameter
	 * @throws NotHomogeneous if queuing system is not homogeneous.
	 */
	public String getOsVersion() throws NoSuchParamException, UnknownParameter, NotHomogeneous;

	/**
	 * If queuing system is homogeneous, this method should return operating system type, common
	 * for all computing resources.
	 * @return operating system type.
	 * @throws NoSuchParamException if parameter is not defined.
	 * @throws UnknownParameter
	 * @throws NotHomogeneous if queuing system is not homogeneous.
	 */
	public String getOsType() throws NoSuchParamException, UnknownParameter, NotHomogeneous;

	/**
	 * If queuing system is homogeneous, this method should return cpu architecture, common for
	 * all computing resources. 
	 * @return cpu architecture.
	 * @throws NoSuchParamException if parameter is not defined.
	 * @throws UnknownParameter
	 * @throws NotHomogeneous if queuing system is not homogeneous.
	 */
	public String getCpuArch() throws NoSuchParamException, UnknownParameter, NotHomogeneous;
	
	/**
	 * If queuing system is homogeneous, this method should return cpu speed, common for all
	 * computing resources.
	 * @return cpu speed.
	 * @throws NoSuchParamException if parameter is not defined.
	 * @throws UnknownParameter
	 * @throws NotHomogeneous if system is not homogeneous.
	 */
	public double getCpuSpeed() throws NoSuchParamException, UnknownParameter, NotHomogeneous;

	/**
	 * If queuing system is homogeneous, this method should return network bandwidth, common
	 * for all computing resources.
	 * @return network bandwidth.
	 * @throws NoSuchParamException if parameter is not defined.
	 * @throws UnknownParameter
	 * @throws NotHomogeneous if queuing resource is not homogeneous.
	 */
	public double getBandwidth() throws NoSuchParamException, UnknownParameter, NotHomogeneous;
	
	/**
	 * If queuing system is homogeneous, this method should return network latency, common
	 * for all computing resources.
	 * @return network latency.
	 * @throws NoSuchParamException if parameter is not defined.
	 * @throws UnknownParameter
	 * @throws NotHomogeneous if queuing system is not homogeneous.
	 */
	public double getLatency() throws NoSuchParamException, UnknownParameter, NotHomogeneous;

	/**
	 * If queuing system is homogeneous, this method should return network capacity, common
	 * for all computing resources.
	 * @return network capacity.
	 * @throws NoSuchParamException if parameter is not defined.
	 * @throws UnknownParameter
	 * @throws NotHomogeneous if queuing system is not homogeneous.
	 */
	public double getCapacity() throws NoSuchParamException, UnknownParameter, NotHomogeneous;

	/**
	 * Check if queuing system is homogeneous. 
	 * See implementation description for details about which parameters are used to determine 
	 * queuing system homogeneity.
	 * @return true if system is homogeneous, false otherwise.
	 */
	public boolean isHomogeneous();
	
}
